#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneController.py
 @brief Control the flight of the drone.
 @date $Date$

 @author 株式会社 東日本計算センター

 プロジェクト方針決定後に記載

"""
import sys

# Import RTM module
import RTC
import OpenRTM_aist

import target_gps
import drone_distance_cal

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# constants
BUTTON_01_MASK = 0x0001
BUTTON_02_MASK = 0x0002
BUTTON_03_MASK = 0x0004
BUTTON_04_MASK = 0x0008
BUTTON_05_MASK = 0x0010
BUTTON_06_MASK = 0x0020
BUTTON_07_MASK = 0x0040
BUTTON_08_MASK = 0x0080
BUTTON_09_MASK = 0x0100
BUTTON_10_MASK = 0x0200
BUTTON_11_MASK = 0x0400
BUTTON_12_MASK = 0x0800
BUTTON_13_MASK = 0x1000
BUTTON_14_MASK = 0x2000

ALLOWABLE_ERROR_POS_CMP = 2.0

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
dronecontroller_spec = ["implementation_id", "DroneController",
         "type_name",         "DroneController",
         "description",       "Control the flight of the drone.",
         "version",           "1.0",
         "vendor",            "EAC",
         "category",          "Drone",
         "activity_type",     "STATIC",
         "max_instance",      "1",
         "language",          "Python",
         "lang_type",         "SCRIPT",
         "conf.default.AltCoefficient", "1",
         "conf.default.LatLonCoefficient", "2",
         "conf.default.LatDiff10m", "0.000137",
         "conf.default.LonDiff10m", "0.000108",
         "conf.default.MoveSpeed", "3.0",
         "conf.__widget__.AltCoefficient", "spin",
         "conf.__widget__.LatLonCoefficient", "spin",
         "conf.__widget__.LatDiff10m", "spin",
         "conf.__widget__.LonDiff10m", "spin",
         "conf.__widget__.MoveSpeed", "spin",
         "conf.__constraints__.AltCoefficient", "1<=x<=10",
         "conf.__constraints__.LatLonCoefficient", "2<=x<=10",
         "conf.__constraints__.LatDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.LonDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.MoveSpeed", "1.0<=x<=10.0",
         ""]
# </rtc-template>


##
# @class DroneController
# @brief Control the flight of the drone.
#
# 別紙"機能仕様書_DroneController"参照
#
# 別紙"機能仕様書_DroneController"参照
#
# 別紙"機能仕様書_DroneController"参照
#
# 別紙"機能仕様書_DroneController"参照
#
#
class DroneController(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_HomeGPSIn = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                        0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ホーム（基地局）GPSデータ
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._HomeGPSInIn = OpenRTM_aist.InPort("HomeGPSIn", self._d_HomeGPSIn)
        self._d_DroneGPSIn = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ドローンGPSデータ
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._DroneGPSInIn = OpenRTM_aist.InPort("DroneGPSIn",
                                                 self._d_DroneGPSIn)
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)
        """
        PS4コントローラのボタン入力状態
         - Type: RTC::TimedULong
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._ButtonInIn = OpenRTM_aist.InPort("ButtonIn", self._d_ButtonIn)
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])
        """
        PS4コントローラのアナログ入力状態
         - Type: RTC::TimedDoubleSeq
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._AnalogInIn = OpenRTM_aist.InPort("AnalogIn", self._d_AnalogIn)
        self._d_TargetGPSOut = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                           0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                           0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        移動ターゲットGPSデータ
         - Type: RTC::GPSTime
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._TargetGPSOutOut = OpenRTM_aist.OutPort("TargetGPSOut",
                                                     self._d_TargetGPSOut)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        移動ターゲット高度算出係数
         - Name: AltCoefficient AltCoefficient
         - DefaultValue: 1
         - Range: 1<=x<=10
        """
        self._AltCoefficient = [1]
        """
        移動ターゲット経緯度算出係数
         - Name: LatLonCoefficient LatLonCoefficient
         - DefaultValue: 2
         - Range: 2<=x<=10
        """
        self._LatLonCoefficient = [2]
        """
        10mあたりの緯度差分
         - Name: LatDiff10m LatDiff10m
         - DefaultValue: 0.000137
         - Range: 0.000000<x<0.100000
        """
        self._LatDiff10m = [0.000137]
        """
        10mあたりの経度差分
         - Name: LonDiff10m LonDiff10m
         - DefaultValue: 0.000108
         - Range: 0.000000<x<0.100000
        """
        self._LonDiff10m = [0.000108]
        """
        ドローン移動速度
         - Name: Speed MoveSpeed
         - DefaultValue: 3.0
         - Range: 1.0<=x<=10.0
        """
        self._MoveSpeed = [3.0]

        # </rtc-template>

        self._target_gps = target_gps.TargetGPS()

        self._home_gps = False
        self._drone_gps = False
        self._button = False
        self._analog = False

        # connection check for data port ButtonInIn
        func = lambda: self._cb_connect_status_change("ButtonIn", True)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("ButtonIn", False)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnDisconnected(ob)

        # connection check for data port AnalogInIn
        func = lambda: self._cb_connect_status_change("AnalogIn", True)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("AnalogIn", False)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnDisconnected(ob)

    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("AltCoefficient", self._AltCoefficient, "1")
        self.bindParameter("LatLonCoefficient", self._LatLonCoefficient, "2")
        self.bindParameter("LatDiff10m", self._LatDiff10m, "0.000137")
        self.bindParameter("LonDiff10m", self._LonDiff10m, "0.000108")
        self.bindParameter("MoveSpeed", self._MoveSpeed, "3.0")

        # Set InPort buffers
        self.addInPort("HomeGPSIn", self._HomeGPSInIn)
        self.addInPort("DroneGPSIn", self._DroneGPSInIn)
        self.addInPort("ButtonIn", self._ButtonInIn)
        self.addInPort("AnalogIn", self._AnalogInIn)

        # Set OutPort buffers
        self.addOutPort("TargetGPSOut", self._TargetGPSOutOut)

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    #
    #    # @return RTC::ReturnCode_t
    #
    #    #
    # def onFinalize(self):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The activated action (Active state entry action)
    #    # former rtc_active_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onActivated(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The deactivated action (Active state exit action)
    #    # former rtc_active_exit()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onDeactivated(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        # PS4キー入力、ドローンステータス取得、およびコントロールデータ送信の周期処理。
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
        self._read_button()
        self._read_analog()
        self._read_home_gps()
        self._read_drone_gps()

        self._set_target_gps()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onError(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    # def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK

    def _read_button(self):
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)

        if self._button:
            self._d_ButtonIn = self._ButtonInIn.read()
            logger.debug("Button = 0b" + format(self._d_ButtonIn.data, '016b'))
        else:
            logger.error("ERROR: ButtonIn is NOT connected")

    def _read_analog(self):
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])

        if self._analog:
            self._d_AnalogIn = self._AnalogInIn.read()
            for i in range(0, len(self._d_AnalogIn.data), 2):
                d1 = d2 = None
                try:
                    d1 = self._d_AnalogIn.data[i]
                    d2 = self._d_AnalogIn.data[i+1]
                except IndexError:
                    pass

                if d2 is None:
                    logger.debug("Analog[%d]:  % f", i, d1)
                else:
                    logger.debug("Analog[%d,%d]: % f % f", i, i+1, d1, d2)
        else:
            logger.error("ERROR: AnalogIn is NOT connected")

    def _read_home_gps(self):
        self._home_gps = False

        if self._HomeGPSInIn.isNew():
            self._d_HomeGPSIn = self._HomeGPSInIn.read()
            self._home_gps = True
            logger.debug("HomeGPS:   % f % f % f",
                         self._d_HomeGPSIn.latitude,
                         self._d_HomeGPSIn.longitude,
                         self._d_HomeGPSIn.altitude)

    def _read_drone_gps(self):
        self._drone_gps = False

        if self._DroneGPSInIn.isNew():
            self._d_DroneGPSIn = self._DroneGPSInIn.read()
            self._drone_gps = True
            logger.debug("DroneGPS:  % f % f % f",
                         self._d_DroneGPSIn.latitude,
                         self._d_DroneGPSIn.longitude,
                         self._d_DroneGPSIn.altitude)

    def _config_target_gps(self):
        config = {"AltCoefficient": self._AltCoefficient[0],
                  "LatLonCoefficient": self._LatLonCoefficient[0],
                  "LatDiff10m": self._LatDiff10m[0],
                  "LonDiff10m": self._LonDiff10m[0],
                  "MoveSpeed": self._MoveSpeed[0]
                  }
        self._target_gps.config_param(config)

    def _set_target_gps(self):
        if not self._home_gps & self._drone_gps:
            logger.error("ERROR: GPS data is NOT available")
            return

        self._config_target_gps()

        target = None
        takeoff_landing = self._takeoff_landing()
        if takeoff_landing:
            target = self._target_gps. \
                get_takeoff_landing_data(self._d_HomeGPSIn, takeoff_landing)
        else:
            mv = self._moving()
            target = self._target_gps.get_move_target_data(self._d_DroneGPSIn,
                                                           mv.front_back,
                                                           mv.right_left,
                                                           mv.up_down)

        self._d_TargetGPSOut.latitude = target.latitude
        self._d_TargetGPSOut.longitude = target.longitude
        self._d_TargetGPSOut.altitude = target.altitude
        self._d_TargetGPSOut.horizontalSpeed = self._MoveSpeed[0]

        logger.debug("TargetGPS: % f % f % f",
                     self._d_TargetGPSOut.latitude,
                     self._d_TargetGPSOut.longitude,
                     self._d_TargetGPSOut.altitude)

        self._TargetGPSOutOut.write()

    def _check_drone_on_home(self):
        ret = False

        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(self._d_HomeGPSIn.longitude,
                                      self._d_DroneGPSIn.longitude,
                                      self._d_HomeGPSIn.latitude,
                                      self._d_DroneGPSIn.latitude,
                                      0,
                                      0)

        # TODO: 最適な誤差をシステムテストで割り出す
        if (dist <= ALLOWABLE_ERROR_POS_CMP and
                self._d_DroneGPSIn.altitude < target_gps.MIN_ALT):

            ret = True

        return ret

    def _takeoff_landing(self):
        operation = 0
        home = self._check_drone_on_home()
        if home:
            # check takeoff request
            if self._d_ButtonIn.data & BUTTON_06_MASK:
                operation = 1
            elif self._d_ButtonIn.data & BUTTON_08_MASK:
                logger.warning("WARNING: Landing operation is "
                               "NOT available "
                               "in the current drone position")
        else:
            # check landing request
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                operation = -1
            elif self._d_ButtonIn.data & BUTTON_06_MASK:
                logger.warning("WARNING: Take-off operation is "
                               "NOT available in the current drone position")

        return operation

    def _moving(self):
        mv1 = self._front_back()
        mv2 = self._right_left()
        mv3 = self._up_down()

        home = self._check_drone_on_home()
        if home:
            if mv1 or mv2 or mv3:
                mv1 = mv2 = mv3 = 0
                logger.warning("WARNING: Move operations are "
                               "NOT available in the current drone position")

        return self.MovingOperation(mv1, mv2, mv3)

    def _front_back(self):
        key = 0
        try:
            key = self._d_AnalogIn.data[0]
        except IndexError:
            pass

        check = lambda x: x > 1 and 1 or x < -1 and -1 or x
        operation = check(key)
        operation *= -1

        return operation

    def _right_left(self):
        operation = 0
        both_push = BUTTON_03_MASK | BUTTON_01_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [right/left]")
        elif self._d_ButtonIn.data & BUTTON_03_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_01_MASK:
            operation = -1

        return operation

    def _up_down(self):
        operation = 0
        both_push = BUTTON_04_MASK | BUTTON_02_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [up/down]")
        elif self._d_ButtonIn.data & BUTTON_04_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_02_MASK:
            operation = -1

        return operation

    def _cb_connect_status_change(self, port, connect):
        if port == "ButtonIn":
            self._button = connect
        elif port == "AnalogIn":
            self._analog = connect
        else:
            logger.warning("Invalid port name: %s", port)

        if not connect:
            logger.warning("ERROR: Data port was disconnected: %s", port)

    class MovingOperation(object):
        def __init__(self, front_back, right_left, up_down):
            self.front_back = front_back
            self.right_left = right_left
            self.up_down = up_down


class ConnectionCallback(OpenRTM_aist.PortCallBack.ConnectionCallback):
    def __init__(self, func):
        self._plugin_func = func

    def __call__(self, profile):
        self._plugin_func()


def DroneControllerInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=dronecontroller_spec)
    manager.registerFactory(profile,
                            DroneController,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    DroneControllerInit(manager)

    # Create a component
    manager.createComponent("DroneController")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()
